/*
** BeNES - Nintendo Entertaiment System Emulator for BeOS
**
** * 1999 by makoto yamagata
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
#include <storage/FilePanel.h>
#include <storage/Path.h>
#include <interface/Alert.h>
#include <interface/MenuItem.h>
#include <stdio.h>
#include "BeNESApp.h"
#include "BeNESWin.h"
#include "BeNESMsg.h"
#include "BeNESJoypad.h"
#include "NesException.h"
#include "NesMainboard.h"
#include "BeFunctions.h"

#define INFO_FORM "\
BeNES Version 0.4.0\n\
\n\
Mapper: %02d\n\
PRG BANKS: %02ld\n\
CHR BANKS: %02ld\n\
MIRRORING: %s\n\
SAVERAM: %s\n\
TRAINER: %s"

int32 BeNESApp::RunBeNES(void* arg)
{
	NesMainboard* obj = reinterpret_cast<NesMainboard*>(arg);
	obj->run();
	return 0;
}

void BeNESApp::beginThread()
{
	mThread = spawn_thread(BeNESApp::RunBeNES, 
		"BeNES", B_NORMAL_PRIORITY, &mMainboard);
	resume_thread(mThread);
}

void BeNESApp::endThread()
{
	if (mThread) {
		status_t result;
		mMainboard.stop();
		wait_for_thread(mThread, &result);
		kill_thread(mThread);
		mThread = 0;
	}
}

void BeNESApp::ReadyToRun()
{
	mOpenPanel = new BFilePanel;
	mWindow = new BeNESWin;
	mJoypad = new BeNESJoypad;
	mMainboard.connectDisplay(mWindow);
	mMainboard.connectJoypad(mJoypad);
	setSkipFrames(2);
	setFPS(60);
	setSound(true);
	mWindow->Show();
}

void BeNESApp::RefsReceived(BMessage* msg)
{
	BMenuItem* infoItem = mWindow->getMenubar()->FindItem(BENES_INFO);
	infoItem->SetEnabled(false);

	entry_ref ref;
	msg->FindRef("refs", 0, &ref);
	
	BEntry entry(&ref);
	BPath path;
	entry.GetPath(&path);
	
	try {
		endThread();
		mWindow->clear();
		mROM.loadRomImage(path.Path());
		infoItem->SetEnabled(true);
		mMainboard.insertROM(&mROM);
		beginThread();
	}
	catch (std::bad_alloc& x) {
		BeShowMessage("cant alloc memory...");
	}
	catch (Nes::FileOpenError& x) {
		BeShowMessage("cant open file...");
	}
	catch (Nes::UnknownFormat& x) {
		BeShowMessage("unknown rom format...");
	}
	catch (Nes::UnknownMapper& x) {
		char buff[256];
		sprintf(buff, "unknown mapper = %d", x.mapper());
		BeShowMessage(buff);
	}

}

bool BeNESApp::QuitRequested()
{
	endThread();
	return true;
}

void BeNESApp::MessageReceived(BMessage* msg)
{
	switch (msg->what) {
	case BENES_OPEN:
		mOpenPanel->Show();
		break;
	case BENES_INFO:
		showRomInfo();
		break;
	case BENES_FRAMESKIP0: setSkipFrames(0); break;
	case BENES_FRAMESKIP1: setSkipFrames(1); break;
	case BENES_FRAMESKIP2: setSkipFrames(2); break;
	case BENES_FRAMESKIP3: setSkipFrames(3); break;
	case BENES_FRAMESKIP4: setSkipFrames(4); break;
	case BENES_FRAMES30: setFPS(30); break;
	case BENES_FRAMES40: setFPS(40); break;
	case BENES_FRAMES50: setFPS(50); break;
	case BENES_FRAMES60: setFPS(60); break;
	case BENES_FRAMESXX: setFPS(00); break;
	case BENES_ENABLESND: setSound(true); break;
	case BENES_DISABLESND: setSound(false); break;
	default:
		BApplication::MessageReceived(msg);
		break;
	}
}

void BeNESApp::showRomInfo()
{
	char buff[512];
	sprintf(buff, INFO_FORM, 
    		mROM.getMapperNumber(),
		mROM.sizeofPRGBANKS(),
		mROM.sizeofCHRBANKS(),
		mROM.isVerticalMirror()? "VERTICAL": "HORIZONTAL",
		mROM.hasSaveRam()? "YES":"NO",
		mROM.hasTrainer()? "YES":"NO");
	BeShowMessage(buff);
}

void BeNESApp::setSkipFrames(int skip)
{
	static uint32 cmdID[5] = {
		BENES_FRAMESKIP0,
		BENES_FRAMESKIP1,
		BENES_FRAMESKIP2,
		BENES_FRAMESKIP3,
		BENES_FRAMESKIP4,
	};
	BMenu* menu = mWindow->getSkipFramesMenu();
	BMenuItem* item = menu->FindMarked();
	if (item) item->SetMarked(false);
	item = mWindow->getMenubar()->FindItem(cmdID[skip]);
	if (item) item->SetMarked(true);
	
	mMainboard.setSkipFrames(skip);
}

void BeNESApp::setFPS(int fps)
{
	BMenu* menu = mWindow->getFramesParSecMenu();
	BMenuItem* item = menu->FindMarked();
	if (item) item->SetMarked(false);
	int32 cmdID;
	switch (fps) {
	case 30: cmdID = BENES_FRAMES30; break;
	case 40: cmdID = BENES_FRAMES40; break;
	case 50: cmdID = BENES_FRAMES50; break;
	case 60: cmdID = BENES_FRAMES60; break;
	default: cmdID = BENES_FRAMESXX; break;
	}
	item = mWindow->getMenubar()->FindItem(cmdID);
	if (item) item->SetMarked(true);

	mWindow->setFPS(fps);
}

void BeNESApp::setSound(bool on)
{
	BMenu* menu = mWindow->getSoundMenu();
	BMenuItem* item = menu->FindMarked();
	if (item) item->SetMarked(false);
	int32 cmdID;
	if (on) {
		cmdID = BENES_ENABLESND; 
	}
	else {
		cmdID = BENES_DISABLESND; 
	}
	item = mWindow->getMenubar()->FindItem(cmdID);
	if (item) item->SetMarked(true);

	mMainboard.enableSound(on);
}
