/*
** BeNES - Nintendo Entertaiment System Emulator for BeOS
**
** * 1999 by makoto yamagata
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
//
// PPU Register Access Wrapper
//
#ifndef PPUREGISTERS_H_DEFINED
#define PPUREGISTERS_H_DEFINED

#include "NesTypes.h"

class NesCPU;

//=============================================================
// PPU 制御レジスタ A
class NesPPUControlA {
	uchar mRegister;
	uchar mLatch;
public:
	NesPPUControlA(): mRegister(0), mLatch(0) { ; }
	// Write Value
	void operator=(uchar val) 
		{ mRegister = val; }
	// VBlank中にNMIを実行するか？
	bool executeNMIonVBlank() {
		return ((mRegister&0x80) != 0); }
	// Spriteサイズを得る
	bool is8x8Sprite()
		{ return ((mRegister&0x20)? false: true); }
	// Background Pattern Table Address(VRAM)を得る
	ushort screenPatternTableAddress()
		{ return ((mRegister&0x10)<<8); }
	// Sprite Pattern Table Address(VRAM)を得る
	ushort spritePatternTableAddress()
		{ return ((mRegister&0x08)<<9); }	
	// PPU Address Increment Valueを得る
	ushort ppuAddressIncrementValue()
		{ return ((mRegister&0x04)? 32: 1); }
	// Name Table Address を得る
	ushort nameTableAddress()
		{ return (0x2000|((mRegister&0x01)<<10)|((mLatch&0x02)<<10)); }
	// Sync Register(call from scanline 0)
	void syncRegister()
		{ mLatch = mRegister; }
	// $2006の書込みは一部をMirror
	void mirrorBit(ushort val);
};

inline void NesPPUControlA::mirrorBit(ushort val)
{
	mRegister &= ~0x03;
	mRegister |= ((val>>10)&0x03);
	syncRegister();
}
//=============================================================
// PPU 制御レジスタ B
class NesPPUControlB {
	uchar mRegister;
public:
	NesPPUControlB(): mRegister(0) { ; }
	// Write Value
	void operator=(uchar val) { mRegister = val; }
	// Sprite Visibility
	bool isSpriteVisible() { return ((mRegister&0x10) != 0); }
	bool isBgVisible() { return ((mRegister&0x08) != 0); }
	bool isGraphicVisible()
		{ return ((mRegister&0x18) != 0); }
};

//=============================================================
// PPU ステータスレジスタ
class NesPPUStatus {
	uchar mRegister;
public:
	NesPPUStatus(): mRegister(0) { ; }
	// Read Value
	uchar operator()() { return mRegister; }
	// VBlank Occurance
	bool inVBlank() { return ((mRegister&0x80) != 0); }
	// Sprite Hit0 Occurance
	bool isHitSprite0() { return ((mRegister&0x40) != 0); }
	// Scanline Sprite Count Max
	bool spriteCountMax() { return ((mRegister&0x20) != 0); }
	// VRAM Write Flag
	bool vramWriteIgnore() { return ((mRegister&0x10) != 0); }
	// VBlank ON/OFF
	void setVBlankBit(bool flag);
	// Sprite Hit ON/OFF
	void setSpriteHitBit(bool flag);
};

inline void NesPPUStatus::setVBlankBit(bool flag)
{
	static const uchar vblankBit = 0x80;
	if (flag) { // ON
		mRegister |= vblankBit;
	}
	else { // OFF
		mRegister &= ~vblankBit;
	}
}

inline void NesPPUStatus::setSpriteHitBit(bool flag)
{
	static const uchar spriteBit = 0x40;
	if (flag) { // ON
		mRegister |= spriteBit;
	}
	else { // OFF
		mRegister &= ~spriteBit;
	}
}

//=============================================================
// PPU SpriteRamAddressレジスタ
class NesPPUSpriteRamAddress {
	uchar mRegister;
public:
	NesPPUSpriteRamAddress(): mRegister(0) { ; }
	// Read Value
	ushort operator()()
		{ return static_cast<ushort>(mRegister); }
	// Write Value
	void operator=(uchar val)
		{ mRegister = val; }
	// Inc Address
	void operator+=(int val)
		{ mRegister=(mRegister+val)&0xff; }
};

//=============================================================
// PPU NesBackGroundScrollレジスタ
class NesPPUBackgroundScroll {
	uchar mRegister;
	uchar  mHScrollValue;
	uchar  mVScrollValue;
	uchar  mVScrollValueLatch;
	bool   mToggle;
public:
	NesPPUBackgroundScroll():
		mRegister(0),
		mHScrollValue(0),
		mVScrollValue(0),
		mVScrollValueLatch(0),
		mToggle(false) { ; }
	// Write Value
	void operator=(uchar value);
	// Get Horizontal Scroll Value
	uchar getHScrollValue() { return mHScrollValue; }
	// Get Vertical Scroll Value
	uchar getVScrollValue() { return mVScrollValue; }
	// Reset H/V Toggle
	void resetToggle() { mToggle = false; }
	void mirrorBit(ushort val, ushort scanline);
	void syncRegister() { mVScrollValue = mVScrollValueLatch; }
	uchar getVScrollLatchValue()
		{ return mVScrollValueLatch; }
};

inline void NesPPUBackgroundScroll::operator=(uchar value)
{
	if (mToggle) { // false は水平 trueは垂直スクロールの値を書く
		mVScrollValueLatch = 0;
		if (value < 240) {
			mVScrollValueLatch = value;
		}
	}
	else {
		mHScrollValue = value;
	}
	mToggle = !mToggle;
}		

inline void NesPPUBackgroundScroll::mirrorBit(ushort val, ushort scanline)
{
	mHScrollValue &= ~0xf8;
 	mHScrollValue |= ((val<<3)&0xf8);
 	mVScrollValue = 0;
	mVScrollValue &= ~0xf8;
	mVScrollValue |= ((val>>2)&0xf8);
	mVScrollValue &= ~0x03;
	mVScrollValue |= ((val>>12)&0x03);
	mVScrollValue -= scanline;
	if (mVScrollValue >= 240) {
		mVScrollValue = 0;
	}
}

//=============================================================
// PPURam Addressレジスタ
class NesPPURamAddress {
	uchar mRegister;
	ushort mAddress;
	bool mToggle;
public:
	NesPPURamAddress():
		mRegister(0),
		mAddress(0),
		mToggle(false) { ; }
	// Write Value
	void operator=(uchar value);
	// Read Value
	ushort operator()()
		{ return mAddress; }
	// Inc Address
	void operator+=(short val)
		{ mAddress+=val; }
	void resetToggle()
		{ mToggle = false; }
};

inline void NesPPURamAddress::operator=(uchar value)
{
	if (mToggle) {
		mAddress = (((mRegister<<8)&0xff00)|(value&0x00ff));
	}
	else {
		mRegister = value;
	}
	mToggle = !mToggle;
}

#endif
