/*
** BeNES - Nintendo Entertaiment System Emulator for BeOS
**
** * 1999 by makoto yamagata
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
//
// ROM IO
//
#include "NesROM.h"
#include "NesException.h"
#include "NesMapperFactory.h"
#include <memory>

//------------------------------------------------------------------------
void NesROM::loadRomImage(const char* name)
{
	// Read ROM Image.
	// Read Header.
	ifstream ifs(name, ios::in|ios::binary);
	if (!ifs.is_open()) { throw Nes::FileOpenError(); }
	ifs.read(reinterpret_cast<char*>(&mHeader), sizeof(mHeader));
	if (memcmp(mHeader.id, "NES\x1a", 4) != 0) {
		throw Nes::UnknownFormat();
	}
	// Create Mapper.
	if (mMapper) { delete mMapper; }
	mMapper = NesMapperFactory::Instance()->Create(getMapperNumber());
	if (mMapper == NULL) {
		throw Nes::UnknownMapper(getMapperNumber());
	}

	// Read Trainer
	auto_ptr<uchar> trainer(readTrainer(ifs));
	// Read PRG-ROM
	int prg_size;
	auto_ptr<uchar> prg(readPrgrom(ifs, &prg_size));
	// Read CHR-ROM
	int chr_size;
	auto_ptr<uchar> chr(readChrrom(ifs, &chr_size));
	// Read SaveRAM
	auto_ptr<uchar> saveram(readSaveRam(name));
	
	// Free Allocated Memories.
	if (mPrgrom) { delete[] mPrgrom; mPrgrom=NULL; }
	if (mChrrom) { delete[] mChrrom; mChrrom=NULL; }
	if (mTrainer) { delete[] mTrainer; mTrainer=NULL; }
	if (mSaveRam) { delete[] mSaveRam; mSaveRam = NULL; }
	
	mPrgromSize = prg_size;
	mChrromSize = chr_size;
	
	// Release auto_ptr.(Keep to Allocated Memory)
	mPrgrom = prg.release();
	mChrrom = chr.release();
	mTrainer = trainer.release();
	mSaveRam = saveram.release();

}

//------------------------------------------------------------------------
NesROM::~NesROM()
{
    delete[] mPrgrom;
    delete[] mChrrom;
    delete[] mTrainer;
    delete[] mSaveRam;
    delete mMapper;
}

//------------------------------------------------------------------------
void NesROM::init(NesIOInterface* io)
{
	// Standard Initiaization.
	NesCPUInterface* cpu = io->getCPU();
	NesPPUInterface* ppu = io->getPPU();
	// Copy Trainer
	if (getTRAINER())
		cpu->memmove(0x7000, getTRAINER(), NES_TRAINER_SIZE);
	// Copy SaveRAM
	if (getSAVERAM())
		cpu->memmove(0x6000, getSAVERAM(), NES_SAVERAM_SIZE);
	// Copy PRGROM
	cpu->memmove(0x8000, getPRGROM(), NES_PRGBANK_SIZE);
	if (sizeofPRGBANKS() == 1) {
		cpu->memmove(0xC000, getPRGROM(), NES_PRGBANK_SIZE);
	}
	else {
		uchar* mem = getPRGROM(sizeofPRGROM()-NES_PRGBANK_SIZE);
		cpu->memmove(0xC000, mem, NES_PRGBANK_SIZE);
	}
	// Copy CHRROM
	ppu->memmove(0x0000, getCHRROM(), NES_CHRBANK_SIZE);
	// Special Initialize.
	mMapper->init(io);
}

//------------------------------------------------------------------------
void NesROM::uninit(NesIOInterface* io)
{
	if (!mSaveName.empty()) {
		ofstream ofs(mSaveName.c_str(), ios::out|ios::binary);
		char* p = reinterpret_cast<char*>(io->getCPU()->getSAVERAM());
		ofs.write(p, NES_SAVERAM_SIZE); 
	}
}

//------------------------------------------------------------------------
uchar* NesROM::readTrainer(ifstream& ifs)
{
	uchar* trainer = NULL;
	if (hasTrainer()) {
		size_t size = NES_TRAINER_SIZE;
		trainer = new uchar[size];
		ifs.read(reinterpret_cast<char*>(trainer), size);
	}
	return trainer;
}

//------------------------------------------------------------------------
uchar* NesROM::readPrgrom(ifstream& ifs, int* prg_size)
{
	*prg_size = sizeofPRGBANKS() * NES_PRGBANK_SIZE;
	uchar* prg = NULL;
	if (*prg_size) {
		prg = new uchar[*prg_size];
		ifs.read(reinterpret_cast<char*>(prg), *prg_size);
	}
	return prg;
}

//------------------------------------------------------------------------
uchar* NesROM::readChrrom(ifstream& ifs, int* chr_size)
{
	*chr_size = sizeofCHRBANKS() * NES_CHRBANK_SIZE;
	uchar* chr = NULL;
	if (*chr_size) {
		chr = new uchar[*chr_size];
		ifs.read(reinterpret_cast<char*>(chr), *chr_size);
	}
	return chr;
}

//------------------------------------------------------------------------
uchar* NesROM::readSaveRam(const char* name)
{
	uchar* saveram = NULL;
	mSaveName.erase();
	if (hasSaveRam()) {
		mSaveName = name;
		memcpy(&(mSaveName[mSaveName.length()-4]), ".sav", 4);
		ifstream ifs(mSaveName.c_str(), ios::in|ios::binary);
		size_t size = NES_SAVERAM_SIZE;
		saveram = new uchar[size];
		if (ifs.is_open()) {
			ifs.read(reinterpret_cast<char*>(saveram), size);
		}
		else {
			memset(saveram, 0, size);
		}
	}
	return saveram;
}
